﻿//     Copyright (c) Microsoft Corporation.  All rights reserved.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace Microsoft.Gateway.Configuration
{
    public static class ConfigurationLoader
    {
        public static RulesConfiguration ReadFromStorage(string indexLocation)
        {
            if (String.IsNullOrWhiteSpace(indexLocation))
            {
                throw new ArgumentNullException("indexLocation", "No value specified for stored configuration index location");
            }
            Uri indexUri;
            if (!Uri.TryCreate(indexLocation, UriKind.Absolute, out indexUri))
            {
                throw new ArgumentException(String.Format("Specified configuration index location is invalid: [{0}]", indexLocation));
            }
            GatewayConfigurationIndex parsedIndex = null;
            try
            {
                parsedIndex = ReadObjectFromBlob<GatewayConfigurationIndex>(indexUri);
            }
            catch (Exception ex)
            {
                throw new InvalidOperationException("Failed to read remote configuration index.", ex);
            }
            var retval = new RulesConfiguration();
            foreach (var roleIndex in parsedIndex.Roles)
            {
                Uri roleConfigUri;
                if (!Uri.TryCreate(roleIndex.Value, UriKind.Absolute, out roleConfigUri))
                {
                    Trace.TraceWarning("Index configuration for role [{0}] does not point to a valid location [{1}]. This role will not be processed.",
                        roleIndex.Key, roleIndex.Value);
                }
                else
                {
                    try
                    {
                        var roleConfig = ReadObjectFromBlob<Role>(roleConfigUri);
                        roleConfig.Segment = roleIndex.Key;
                        retval.Roles[roleIndex.Key] = roleConfig;
                    }
                    catch 
                    {
                        // Details of exception have already been logged
                        Trace.TraceWarning("Failed to read role configuration for role [{0}]. This role will not be processed.", roleIndex.Key);
                    }
                }
            }
            retval.Directory = parsedIndex.Directory;
            retval.ApplicationName = parsedIndex.ApplicationName;
            return retval;
        }

        public static RulesConfiguration ReadFromCache(string cachePath)
        {
            if (!File.Exists(cachePath))
            {
                return null;
            }
            try
            {
                using (var reader = new JsonTextReader(File.OpenText(cachePath)))
                {
                    return new JsonSerializer().Deserialize<RulesConfiguration>(reader);
                }
            }
            catch (Exception ex)
            {
                Trace.TraceWarning("Failure writing local configuration cache file [{0}]. Details: {1}", cachePath, ex);
            }
            return null;
        }

        public static bool WriteToCache(this RulesConfiguration configuration, string cachePath)
        {
            try
            {
                using (var writer = File.CreateText(cachePath))
                {
                    var serializer = new JsonSerializer
                    {
                        Formatting = Formatting.Indented,
                    };
                    serializer.Serialize(writer, configuration);
                    return true;
                }
            }
            catch (Exception ex)
            {
                Trace.TraceWarning("Failure writing local configuration cache file [{0}]. Details: {1}", cachePath, ex);
            }
            return false;
        }

        static JObject ReadJsonFromBlob(Uri blobLocation)
        {
            return ReadJsonFromBlob<JObject>(blobLocation, reader => JObject.Load(reader));
        }

        static T ReadObjectFromBlob<T>(Uri blobLocation)
        {
            return ReadJsonFromBlob<T>(blobLocation, reader => new JsonSerializer().Deserialize<T>(reader));
        }

        static T ReadJsonFromBlob<T>(Uri blobLocation, Func<JsonReader, T> deserializer)
        {
            try
            {
                var request = (HttpWebRequest)HttpWebRequest.Create(blobLocation);
                using (var response = request.GetResponse() as HttpWebResponse)                
                {
                    return deserializer(new JsonTextReader(new StreamReader(response.GetResponseStream())));
                }
            }
            catch (Exception ex)
            {
                Trace.TraceWarning("Failure reading JSON configuration from blob [{0}]. Details: {1}", blobLocation, ex);
                throw;
            }
        }
    }
}
