﻿//     Copyright (c) Microsoft Corporation.  All rights reserved.

using System;
using System.Diagnostics;
using System.IdentityModel.Tokens;
using System.IO;
using System.Linq;
using System.Xml.Linq;
using System.Xml.XPath;
using Microsoft.Gateway.Utils;
using Microsoft.Web.Administration;

namespace Microsoft.Gateway.RulesConfigurator
{
    class DirectoryConfigurator
    {
        public static bool ApplyConfiguration(string domain, string applicationName)
        {
            if (String.IsNullOrWhiteSpace(domain))
            {
                Trace.TraceInformation("DirectoryConfigurator: directory not set, so not running configuration.");
                return true;
            }
            if (String.IsNullOrWhiteSpace(applicationName))
            {
                Trace.TraceInformation("DirectoryConfigurator: applicationName not set, so not running configuration.");
                return true;
            }
            try
            {
                var gatewaySite = new ServerManager().Sites[AzureUtils.GetDefaultIISSiteName()];
                var configPath = Environment.ExpandEnvironmentVariables(Path.Combine(gatewaySite.Applications.First().VirtualDirectories.First().PhysicalPath, "web.config"));
                var value = XNamespace.None + "value";
                var metadataAddress = @"https://login.windows.net/" + domain + "/federationmetadata/2007-06/federationmetadata.xml";
                var login = @"https://login.windows.net/" + domain + "/wsfed";

                XDocument config = XDocument.Load(configPath);
                var issuingAuthority = ValidatingIssuerNameRegistry.GetIssuingAuthority(metadataAddress);

                config.XPathSelectElement("//configuration/appSettings/add[@key = 'ida:FederationMetadataLocation']")
                    .SetAttributeValue(value, metadataAddress);
                config.XPathSelectElement("//configuration/appSettings/add[@key = 'ida:Issuer']")
                    .SetAttributeValue(value, login);
                config.XPathSelectElement("//configuration/system.identityModel/identityConfiguration/audienceUris/add")
                    .SetAttributeValue(value, applicationName);
                config.XPathSelectElement("//configuration/system.identityModel/identityConfiguration/issuerNameRegistry/authority")
                    .SetAttributeValue("name", issuingAuthority.Name);
                config.XPathSelectElement("//configuration/system.identityModel/identityConfiguration/issuerNameRegistry/authority/keys/add")
                    .SetAttributeValue("thumbprint", issuingAuthority.Thumbprints.First());
                config.XPathSelectElement("//configuration/system.identityModel/identityConfiguration/issuerNameRegistry/authority/validIssuers/add")
                    .SetAttributeValue("name", issuingAuthority.Name);
                config.XPathSelectElement("//configuration/system.identityModel.services/federationConfiguration/wsFederation")
                    .SetAttributeValue("issuer", login);
                config.XPathSelectElement("//configuration/system.identityModel.services/federationConfiguration/wsFederation")
                    .SetAttributeValue("realm", applicationName);
                config.XPathSelectElement("//configuration/system.web/authorization")
                    .ReplaceAll(new XElement(XNamespace.None + "deny", new XAttribute(XNamespace.None + "users", "?")));

                config.Save(configPath);
                return true;
            }
            catch (Exception ex)
            {
                Trace.TraceError("Failed to update directory for web server. Details: {0}", ex);
            }
            return false;
        }
    }
}
