﻿//     Copyright (c) Microsoft Corporation.  All rights reserved.

using System;
using System.Collections.Generic;
using System.Configuration;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using Microsoft.Gateway.Configuration;
using Microsoft.Gateway.Utils;

namespace Microsoft.Gateway.RulesConfigurator
{
    class Program
    {
        const string ConfigKeyCacheLocation = "CacheLocation";
        const string CacheFilename          = "GatewayConfiguration.cache";
        const string LogFilename            = "GatewayConfiguration.log";

        static int Main(string[] args)
        {
            bool addLogListener = false;
            if (!Trace.Listeners.OfType<ConsoleTraceListener>().Any())
            {
                Trace.Listeners.Add(new ConsoleTraceListener());
                addLogListener = true;
            }
            AzureUtils.AddAzureDiagnosticsListner();
            if (args.Length < 1)
            {
                Trace.TraceError("Invalid arguments. Usage: RulesConfiguration {ConfigurationLocation} [ConfigurationCache]");
                return 1;
            }
            string storageLocation = args[0];
            string cacheLocation = null;
            if (args.Length >= 2)
            {
                cacheLocation = args[1];
                // Update the config file so that we don't need to keep specifying this argument
                try
                {
                    var appConfig = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None);
                    var settings = appConfig.AppSettings.Settings;
                    var cacheLocationElement = settings[ConfigKeyCacheLocation];
                    if (cacheLocationElement == null)
                    {
                        settings.Add(ConfigKeyCacheLocation, cacheLocation);
                    }
                    else
                    {
                        cacheLocationElement.Value = cacheLocation;
                    }
                    appConfig.Save(ConfigurationSaveMode.Full);
                }
                catch (Exception ex)
                {
                    Trace.TraceError("Failed to update application configuration. Details: {0}", ex);
                    return 5;
                }
            }
            if (String.IsNullOrWhiteSpace(cacheLocation))
            {
                cacheLocation = Environment.ExpandEnvironmentVariables(GetAppConfig(ConfigKeyCacheLocation, "%temp%"));
            }
            if (String.IsNullOrWhiteSpace(cacheLocation))
            {
                Trace.TraceError("Invalid cache location");
                return 2;
            }
            if (addLogListener)
            {
                Trace.Listeners.Add(new TextWriterTraceListener(Path.Combine(cacheLocation, LogFilename)));
            }
            string cachePath = Path.Combine(cacheLocation, CacheFilename);
            Trace.TraceInformation("RulesConfigurator (Version: {0}): Reading routing rules configuration from storage account [{1}]. Comparing to cached version at [{2}]",
                Assembly.GetEntryAssembly().GetName().Version, storageLocation, cachePath);
            RulesConfiguration downloadedConfiguration = null;
            try
            {
                downloadedConfiguration = ConfigurationLoader.ReadFromStorage(storageLocation);
            }
            catch (Exception ex)
            {
                Trace.TraceError("Failed to retrieve remote Gateway configuration from [{0}]. Details: {1}", storageLocation, ex);
                return 3;
            }
            bool cacheExpired = true;
            if (File.Exists(cachePath))
            {
                try
                {
                    var cachedConfiguration = ConfigurationLoader.ReadFromCache(cachePath);
                    cacheExpired = !downloadedConfiguration.Equals(cachedConfiguration);
                }
                catch (Exception ex)
                {
                    Trace.TraceWarning("Failed to compare cached configuration. Details: {0}", ex);
                }
            }
            if (cacheExpired)
            {
                if (!WebsiteConfigurator.ApplyConfiguration(downloadedConfiguration) ||
                    !DirectoryConfigurator.ApplyConfiguration(downloadedConfiguration.Directory, downloadedConfiguration.ApplicationName))
                {
                    // Failed already logged, just quit
                    return 4;
                }
                downloadedConfiguration.WriteToCache(cachePath);
                Trace.TraceInformation("Updated configuration has been applied to web site");
            }
            else
            {
                Trace.TraceInformation("Configuration has not be updated. No routing rule changes have been applied.");
            }

            return 0;
        }

        static string GetAppConfig(string configName, string defaultValue)
        {
            string configValue = ConfigurationManager.AppSettings[ConfigKeyCacheLocation];
            if (String.IsNullOrWhiteSpace(configValue))
            {
                return defaultValue;
            }
            return configValue;
        }
    }
}
