﻿//     Copyright (c) Microsoft Corporation.  All rights reserved.

using System;
using System.Diagnostics;
using System.Linq;
using Microsoft.Gateway.Configuration;
using Microsoft.Web.Administration;

namespace Microsoft.Gateway.RulesConfigurator
{
    public class WebsiteConfigurator
    {
        // Server variables
        const string SERVER_VARIABLE_TARGET_URI         = "TARGET_URI";                
        const string SERVER_VARIABLE_CLIENT_REQUEST_ID  = "HTTP_client_request_Id";
        const string SERVER_VARIABLE_REMOTE_ADDR        = "REMOTE_ADDR";        
        const string SERVER_VARIABLE_HTTP_COOKIE        = "HTTP_COOKIE";         
        const string SERVER_VARIABLE_CACHE_URL          = "CACHE_URL";
        const string SERVER_VARIABLE_ACCEPT_ENCODING    = "HTTP_ACCEPT_ENCODING";
        const string SERVER_VARIABLE_ORIGINAL_ACCEPT_ENCODING = "ORIGINAL_HTTP_ACCEPT_ENCODING";
        const string SERVER_VARIABLE_RESPONSE_CONTENT_TYPE = "RESPONSE_CONTENT_TYPE";
        const string SERVER_VARIABLE_URL                = "URL";
        const string SERVER_VARIABLE_RESPONSE_STATUS    = "RESPONSE_STATUS";
        const string SERVER_VARIABLE_RESPONSE_LOCATION  = "RESPONSE_LOCATION";
        const string SERVER_VARIABLE_OUTBOUND_RULE_FILTER = "OUTBOUND_RULE_FILTER";
        const string SERVER_VARIABLE_HTTP_X_REMOTE_USER = "HTTP_X_REMOTE_USER";
        const string SERVER_VARIABLE_HTTP_X_TENANT = "HTTP_X_TENANT";

        // Well-known elements
        const string ELEMENT_ISHTML_PRECONDITION        = "ResponseIsHtml";
        const string ELEMENT_ISREDIRECT_PRECONDITION    = "IsRedirect";
        const string ELEMENT_NEEDS_RESTORE_ACCEPT_ENCODING_PRECONDITION = "NeedsRestoreAcceptEncoding";
        
        // regex generation
        const string SLASH_DOT = @"\.";
        const string REGEX_0_TO_255 = "([0]|[1-9]\\d?|[1]\\d{2}|2([0-4]\\d|5[0-5]))";
        
        static readonly string[] AllowedServerVariables = new[] 
        {
            SERVER_VARIABLE_TARGET_URI,                        
            SERVER_VARIABLE_CLIENT_REQUEST_ID,            
            SERVER_VARIABLE_ACCEPT_ENCODING,
            SERVER_VARIABLE_OUTBOUND_RULE_FILTER,
            SERVER_VARIABLE_HTTP_X_REMOTE_USER,
            SERVER_VARIABLE_HTTP_X_TENANT
            
        };

        public static bool ApplyConfiguration(RulesConfiguration configToApply)
        {
            try
            {
                using (ServerManager serverManager = new ServerManager())
                {
                    var webConfig = serverManager.GetApplicationHostConfiguration();
                    var webFarmsCollection = WebfarmConfigurator.ConfigureWebFarmsNodeDefaults(webConfig);
                    
                    // Make sure our <allowedServerVariables> is complete
                    var allowedVariablesCollection = webConfig.GetSection("system.webServer/rewrite/allowedServerVariables").GetCollection();
                    var existingVariables = allowedVariablesCollection
                        .Select(variableElement => variableElement["name"].ToString())
                        .ToLookup(variable => variable, StringComparer.OrdinalIgnoreCase);
                    foreach (string allowedVariable in AllowedServerVariables)
                    {
                        if (!existingVariables.Contains(allowedVariable))
                        {
                            var variableElement = allowedVariablesCollection.CreateElement("add");
                            variableElement["name"] = allowedVariable;
                            allowedVariablesCollection.Add(variableElement);
                        }
                    }
              
                    // Create the inbound & outbound rules
                    var inboundRulesCollection = webConfig.GetSection("system.webServer/rewrite/globalRules").GetCollection();
                    var outboundRulesSection = webConfig.GetSection("system.webServer/rewrite/outboundRules");
                    var outboundRulesCollection = outboundRulesSection.GetCollection();
                    // Remove all existing rules before adding the new rules
                    inboundRulesCollection.Clear();
                    outboundRulesCollection.Clear();
                    
                    //Add all rules that must be processed first
                    inboundRulesCollection.Add(CreateInboundGwRemoteUserRule(inboundRulesCollection.CreateElement("rule")));

                    // Setup some shared pre-conditions                   
                    var preConditionsCollection = outboundRulesSection.GetCollection("preConditions");
                    preConditionsCollection.Clear();

                    var sharedPreConditions = CreatePrecondition(preConditionsCollection.CreateElement("preCondition"), ELEMENT_ISHTML_PRECONDITION, 
                                        Tuple.Create(MakeVariableReference(SERVER_VARIABLE_RESPONSE_CONTENT_TYPE), @"^text/html"));

                    preConditionsCollection.Add(sharedPreConditions);                    
                                 
                    // Apply role rules in alphabetical order
                    foreach (var role in configToApply.OrderedRoles)
                    {
                        if (role.Item2.Flights.Count > 0)
                        {                           
                            var defaultFlights = role.Item2.OrderedFlights.Where(f => f.Item2.IsDefaultFlight);
                            if(defaultFlights.Any())
                            {
                                //Create the webfarm for this role                                
                                WebfarmConfigurator.CreateWebFarmForRole(webFarmsCollection, role, defaultFlights);

                                //Create one inbound rule for each webfarm created in ARR 
                                inboundRulesCollection.Add(CreateFlightRuleARR(inboundRulesCollection.CreateElement("rule"), role.Item2, role.Item1));

                                //Create outbound rule for webfarm
                                CreateOutboundRulesARR(outboundRulesCollection, preConditionsCollection, role.Item2);
                            }    

                            foreach (var flight in role.Item2.OrderedFlights.Where(f => !f.Item2.IsDefaultFlight))
                            {
                                inboundRulesCollection.Add(CreateFlightRule(
                                    inboundRulesCollection.CreateElement("rule"), role.Item1, role.Item2, flight.Item1, flight.Item2));                                                                
                                CreateOutboundRules(outboundRulesCollection, preConditionsCollection, role.Item1, role.Item2, flight.Item1);
                            }               
                        }
                        else
                        {
                            inboundRulesCollection.Add(CreateRoleRule(
                                inboundRulesCollection.CreateElement("rule"), role.Item1, role.Item2));
                            CreateOutboundRules(outboundRulesCollection, preConditionsCollection, role.Item1, role.Item2);
                        }
                    }

                    CreateRestoreAcceptEncodingRule(outboundRulesCollection, preConditionsCollection);

                    serverManager.CommitChanges();
                }
                return true;
            }
            catch (Exception ex)
            {
                Trace.TraceError("Failed to update routing rules for web server. Details: {0}", ex);
            }
            return false;
        }
      
      
        static ConfigurationElement CreateFlightRule(ConfigurationElement ruleElement, string roleName, Role role, string flightName, Flight flight)
        {
            CreateInboundRule(ruleElement, roleName, role, flightName, flight.GetEffectiveRedirectBase(role),
                conditionsCollection =>
                {
                    if (flight.Pivot != FlightPivots.Default)
                    {
                        var addElement = conditionsCollection.CreateElement("add");
                        addElement["input"] = GetServerVariableForPivot(flight.Pivot);
                        addElement["pattern"] = GetPatternForPivot(flight.Pivot, flight.Value);
                        conditionsCollection.Add(addElement);
                    }
                });
            return ruleElement;
        }


        static ConfigurationElement CreateFlightRuleARR(ConfigurationElement ruleElement, Role role, string arrServerFarmName)
        {
            CreateInboundRule(ruleElement, role.Segment, role, arrServerFarmName, arrServerFarmName, null);
            return ruleElement;
        }
     
        private static string GetPatternForPivot(FlightPivots pivot, string pivotValue)
        {
            switch(pivot)
            {
                case FlightPivots.Default:
                    return string.Empty;

                case FlightPivots.SourceIP:
                    return GetIPRegEx(pivotValue);

                case FlightPivots.Tenant:
                    return GetExactMatchRegEx(pivotValue);

                case FlightPivots.User:
                    return GetExactMatchRegEx(pivotValue);
                
                default:
                    Debug.Assert(false);
                    return String.Empty;                                    
            }
        }

        static ConfigurationElement CreateRoleRule(ConfigurationElement ruleElement, string roleName, Role role)
        {
            CreateInboundRule(ruleElement, roleName, role, "Default", role.RedirectBase, null);
            return ruleElement;
        }

        static void CreateInboundRule(ConfigurationElement ruleElement, 
            string roleSegment, 
            Role role, 
            string flightName,
            string redirectBaseUri,
            Action<ConfigurationElementCollection> addConditions)
        {            
            ruleElement["name"] = String.Format("{0}_{1}", roleSegment, flightName);
            ruleElement["patternSyntax"] = "ECMAScript";
            ruleElement["stopProcessing"] = true;

            string captureGroup = @"{R:2}";
            var matchElement = ruleElement.GetChildElement("match");
            if (!role.IsDefaultRole)
            {
                matchElement["url"] = String.Format("^{0}($|/(.*))", roleSegment);
            }
            else
            {
                // Catch-all rule
                matchElement["url"] = "(.*)";
                captureGroup = @"{R:1}";
            }

            var conditionsElement = ruleElement.GetChildElement("conditions");
            conditionsElement["trackAllCaptures"] = true;
            var conditionsCollection = conditionsElement.GetCollection();
            // Extract the protocol from the request to preserve only if the redirect uri uses it {C:1}
            bool preserveProtocol = false;
            if (!redirectBaseUri.StartsWith("http", StringComparison.OrdinalIgnoreCase))
            {
                var protocolElement = conditionsCollection.CreateElement("add");
                protocolElement["input"] = MakeVariableReference(SERVER_VARIABLE_CACHE_URL);
                protocolElement["pattern"] = "^(https?://)";
                conditionsCollection.Add(protocolElement);
                preserveProtocol = true;
            }
            // Add other conditions
            if (addConditions != null)
            {
                addConditions(conditionsCollection);
            }
            // Set some server variables for timing & rewrite (must have been previously defined in <allowedServerVariables> section)
            var variablesCollection = ruleElement.GetChildElement("serverVariables").GetCollection();
            // TODO: Add in when we have request timing
            //AddServerVariable(variablesCollection, "START_TARGETREQUEST", "{TimeProvider:}");
            string targetUri = (preserveProtocol ? @"{C:1}" : "") + redirectBaseUri + "/" + captureGroup;
            AddServerVariable(variablesCollection, SERVER_VARIABLE_TARGET_URI, targetUri);            
            AddServerVariable(variablesCollection, SERVER_VARIABLE_ORIGINAL_ACCEPT_ENCODING, MakeVariableReference(SERVER_VARIABLE_ACCEPT_ENCODING));
            //Archive HTTP_ACCEPT_ENCODING before clear it out
            AddServerVariable(variablesCollection, SERVER_VARIABLE_ACCEPT_ENCODING, "");
            AddServerVariable(variablesCollection, SERVER_VARIABLE_OUTBOUND_RULE_FILTER, ruleElement["name"].ToString());            
            
            var actionElement = ruleElement.GetChildElement("action");
            actionElement["type"] = "Rewrite";
            actionElement["url"] = MakeVariableReference(SERVER_VARIABLE_TARGET_URI);
        }

        private static ConfigurationElement CreateInboundGwRemoteUserRule(ConfigurationElement ruleElement)
        {
            ruleElement["name"] = "Set User + Tenant";
            ruleElement["patternSyntax"] = "ECMAScript";
            ruleElement["stopProcessing"] = false;

            var matchElement = ruleElement.GetChildElement("match");
            matchElement["url"] = "(.*)";

            var conditionsElement = ruleElement.GetChildElement("conditions");
            conditionsElement["trackAllCaptures"] = true;
            var conditionsCollection = conditionsElement.GetCollection();

            //Get the username from the UPN part of the HTTP_COOKIE
            var protocolElement = conditionsCollection.CreateElement("add");
            protocolElement["input"] = MakeVariableReference(SERVER_VARIABLE_HTTP_COOKIE);
            protocolElement["pattern"] = @"(;| )?UPN=([^@]+@([^;]+))";
            conditionsCollection.Add(protocolElement);

            // set the server variable
            var variablesCollection = ruleElement.GetChildElement("serverVariables").GetCollection();
            AddServerVariable(variablesCollection, SERVER_VARIABLE_HTTP_X_REMOTE_USER, "{C:2}");
            AddServerVariable(variablesCollection, SERVER_VARIABLE_HTTP_X_TENANT, "{C:3}");

            var actionElement = ruleElement.GetChildElement("action");
            actionElement["type"] = "None";

            return ruleElement;
        }

        static void CreateOutboundRulesARR(ConfigurationElementCollection outboundRules, ConfigurationElementCollection preConditions, Role role)
        {
            CreateHtmlOutboundRule(outboundRules, preConditions, role.Segment, role, role.Segment, CreateRewriteOutboundRuleARR);
            CreateRedirectOutboundRule(outboundRules, preConditions, role.Segment, role, role.Segment, CreateRewriteOutboundRuleARR);
        }


        static void CreateOutboundRules(ConfigurationElementCollection outboundRules, ConfigurationElementCollection preConditions, string roleName, Role role, string flightName = null)
        {
            CreateHtmlOutboundRule(outboundRules, preConditions, roleName, role, flightName, CreateRewriteOutboundRule);
            CreateRedirectOutboundRule(outboundRules, preConditions, roleName, role, flightName, CreateRewriteOutboundRule);
        }

        static void CreateHtmlOutboundRule(ConfigurationElementCollection outboundRules, ConfigurationElementCollection preConditions, string roleName, Role role, string flightName
                                            , Func<ConfigurationElement, string, Role, string, string, string, string, ConfigurationElement> CreateRule)
        {
            var baseUri = (!string.IsNullOrWhiteSpace(flightName) && role.Flights.ContainsKey(flightName)) ? role.Flights[flightName].GetEffectiveRedirectBase(role) : role.RedirectBase;
            var hostname = GetHostName(baseUri);
            flightName = flightName ?? "Default";

            // Add a pre-condition for this host
            string preConditionName = String.Format("{0}_{1}_{2}", roleName, flightName, ELEMENT_ISHTML_PRECONDITION);
            preConditions.Add(CreatePrecondition(preConditions.CreateElement("preCondition"), preConditionName,
                Tuple.Create(SERVER_VARIABLE_RESPONSE_CONTENT_TYPE, "^text/html"),                
                Tuple.Create(SERVER_VARIABLE_OUTBOUND_RULE_FILTER, String.Format("^{0}_{1}$", roleName, flightName))));

            // Relative & Absolute Rule
            outboundRules.Add(CreateRule(
                outboundRules.CreateElement("rule"),
                String.Format("{0}_{1}", roleName, flightName),
                role,
                preConditionName,
                "filterByTags",
                "A, Area, Base, Form, Frame, Head, IFrame, Img, Input, Link, Script",
                hostname));
        }

        static void CreateRedirectOutboundRule(ConfigurationElementCollection outboundRules, ConfigurationElementCollection preConditions, string roleName, Role role, string flightName
                                                , Func<ConfigurationElement, string, Role, string, string, string, string, ConfigurationElement> CreateRule)
        {
            var baseUri = (!string.IsNullOrWhiteSpace(flightName) && role.Flights.ContainsKey(flightName)) ? role.Flights[flightName].GetEffectiveRedirectBase(role) : role.RedirectBase;
            var hostname = GetHostName(baseUri);
            flightName = flightName ?? "Default";
            
            string preConditionName = String.Format("{0}_{1}_{2}", roleName, flightName, ELEMENT_ISREDIRECT_PRECONDITION);
            preConditions.Add(CreatePrecondition(preConditions.CreateElement("preCondition"), preConditionName,
                Tuple.Create(SERVER_VARIABLE_RESPONSE_STATUS, @"3\d\d"),                
                Tuple.Create(SERVER_VARIABLE_OUTBOUND_RULE_FILTER, String.Format("^{0}_{1}$", roleName, flightName))));

            // Relative & Absolute Rule
            outboundRules.Add(CreateRule(
                outboundRules.CreateElement("rule"),
                String.Format("{0}_{1}_Redirect", roleName, flightName),
                role,
                preConditionName,
                "serverVariable",
                SERVER_VARIABLE_RESPONSE_LOCATION,
                hostname));
        }

        static void CreateRestoreAcceptEncodingRule(ConfigurationElementCollection outboundRules, ConfigurationElementCollection preConditions)
        {
            preConditions.Add(CreatePrecondition(preConditions.CreateElement("preCondition"), ELEMENT_NEEDS_RESTORE_ACCEPT_ENCODING_PRECONDITION,
                Tuple.Create(SERVER_VARIABLE_ORIGINAL_ACCEPT_ENCODING, @".+")));

            outboundRules.Add(CreateRewriteOutboundRule(
                outboundRules.CreateElement("rule"),
                "_RestoreAcceptEncoding",
                ELEMENT_NEEDS_RESTORE_ACCEPT_ENCODING_PRECONDITION,
                "serverVariable",
                SERVER_VARIABLE_ACCEPT_ENCODING,
                @"(.*)",
                @"Rewrite",
                MakeVariableReference(SERVER_VARIABLE_ORIGINAL_ACCEPT_ENCODING)
                ));
        }

        static ConfigurationElement CreateRewriteOutboundRuleARR(ConfigurationElement rule,
                                                         string ruleName,
                                                         Role role,
                                                         string preConditionName,
                                                         string matchSubject,
                                                         string matchValue,
                                                         string redirectedHostName)
        {
            rule["name"] = ruleName;
            rule["preCondition"] = preConditionName;

            var hostNames = string.Join("|", role.OrderedFlights.Where(f => f.Item2.IsDefaultFlight)
                                        .Select(f => string.Format("https?://{0}", EscapeHostName(f.Item2.GetEffectiveRedirectBase(role)))));

            var matchElement = rule.GetChildElement("match");
            matchElement[matchSubject] = matchValue;
            matchElement["pattern"] = String.Format("^(/(.*)|({0})($|/(.*)|(\\?.*)))", hostNames);

            var action = rule.GetChildElement("action");
            action["type"] = @"Rewrite";
            if (!role.IsDefaultRole)
            {
                action["value"] = @"/" + role.Segment + @"/{R:2}{R:5}{R:6}";
            }
            else
            {
                action["value"] = @"/{R:2}{R:5}{R:6}";
            }

            return rule;
        }


        static ConfigurationElement CreateRewriteOutboundRule(ConfigurationElement rule, 
            string ruleName, 
            Role role, 
            string preConditionName, 
            string matchSubject, 
            string matchValue, 
            string redirectedHostName)
        {
            string matchPattern = String.Format("^(/(.*)|https?://{0}($|/(.*)|(\\?.*)))", EscapeHostName(redirectedHostName));
            string actionValue = "";
            if (!role.IsDefaultRole)
            {
                actionValue = @"/" + role.Segment + @"/{R:2}{R:4}{R:5}";
            }
            else
            {
                actionValue = @"/{R:2}{R:4}{R:5}";
            }

            return CreateRewriteOutboundRule(rule, ruleName, preConditionName, matchSubject, matchValue, matchPattern, @"Rewrite", actionValue);
        }

        static ConfigurationElement CreateRewriteOutboundRule(ConfigurationElement rule, 
            string ruleName, 
            string preConditionName, 
            string matchSubject, 
            string matchValue, 
            string matchPattern,
            string actionType, 
            string actionValue)
        {
            rule["name"] = ruleName;
            rule["preCondition"] = preConditionName;

            var matchElement = rule.GetChildElement("match");
            matchElement[matchSubject] = matchValue;
            matchElement["pattern"] = matchPattern;

            var action = rule.GetChildElement("action");
            action["type"] = actionType;
            action["value"] = actionValue;

            return rule;
        }

        static ConfigurationElement CreatePrecondition(ConfigurationElement preCondition, string conditionName, params Tuple<string, string>[] matchArguments)
        {
            preCondition["name"] = conditionName;
            var preConditionCollection = preCondition.GetCollection();
            foreach (var conditionMatch in matchArguments)
            {
                var matchElement = preConditionCollection.CreateElement("add");
                matchElement["input"] = MakeVariableReference(conditionMatch.Item1);
                matchElement["pattern"] = conditionMatch.Item2;
                preConditionCollection.Add(matchElement);
            }
            return preCondition;
        }

        static string GetHostName(string baseUri)
        {
            Uri hostUri;
            if (Uri.TryCreate(baseUri, UriKind.Absolute, out hostUri))
            {
                return hostUri.Host;
            }
            int pos = baseUri.IndexOf('/');
            if (pos != -1)
            {
                return baseUri.Substring(0, pos - 1);
            }
            return baseUri;
        }

        static string EscapeHostName(string hostName)
        {
            return hostName.Replace(".", @"\.");
        }

        static string GetServerVariableForPivot(FlightPivots pivot)
        {
            switch (pivot)
            {
                case FlightPivots.Default:
                    return String.Empty;

                case FlightPivots.SourceIP:
                    return MakeVariableReference(SERVER_VARIABLE_REMOTE_ADDR);

                case FlightPivots.Tenant:
                    return MakeVariableReference(SERVER_VARIABLE_HTTP_X_TENANT);

                case FlightPivots.User:
                    return MakeVariableReference(SERVER_VARIABLE_HTTP_X_REMOTE_USER);

                default:
                    Debug.Assert(false);
                    return String.Empty;
            }
        }

        static string MakeVariableReference(string variableName)
        {
            return '{' + variableName + '}';
        }

        static void AddServerVariable(ConfigurationElementCollection variablesCollection, string name, string value)
        {
            var variableElement = variablesCollection.CreateElement("set");
            variableElement["name"] = name;
            variableElement["value"] = value;
            variablesCollection.Add(variableElement);
        }

        /// <summary>
        /// Generate a regex to validate the IP given in the Flight Value property                
        /// </summary>
        /// <param name="value">Flight SourceIP pivot value</param>
        /// <returns>string representing a regex for the IP</returns>
        static string GetIPRegEx(string value)
        {
            if (string.IsNullOrWhiteSpace(value))
            {
                throw new ArgumentNullException("Flight Value must be non null, e.g 1.2.3.4 for exact match and 1.2.3. for match on ip range 1.2.3.0 - 1.2.3.255");
            }

            //otherwise generate exact match expression or wildcard
            // e.g input  1.2.3.4 retun a regex with exact match ^1\.2\.3\.4$
            //     input with a missing part requires "wildcard" expression similar to provide 0-255 for that missing part
            //      example input of 1.2.3 will generate regex ^1\.2\.3\.([0]|[1-9]\\d?|[1]\\d{2}|2([0-4]\\d|5[0-5]))$          
            
            var parts = value.Split('.');           
            if(parts.Length < 4)
            {
                Array.Resize(ref parts, 4); //pad the array
            }

            var expression = string.Join(SLASH_DOT, parts.Select(part => string.IsNullOrWhiteSpace(part) ? REGEX_0_TO_255 : part));                
            return string.Format("^{0}$", expression);                                            
        }

        /// <summary>
        /// Generate an exact match regex 
        /// </summary>
        /// <param name="value">Flight Tenant or user pivot value</param>
        /// <returns>RegEx representing the Flight tenant or user pivot value</returns>
        private static string GetExactMatchRegEx(string value)
        {
            if (string.IsNullOrWhiteSpace(value))
            {
                throw new ArgumentNullException("Flight Value must be non null");
            }

            // generate a regex that is used to validate flights using either user or tenant against the 
            // HTTP_X_TENANT and HTTP_X_REMOTE_USER server variables
            // e.g HTTP_X_REMOTE_USER = nickha@microsoft.com resulting regex = (^nickha@microsoft\.com$)
            // if the provided value is a comma separated string e.g nickha@microsoft.com, nickha@bing.com generate (^nickha@microsoft\.com$)|(^nickha@bing\.com$)
            // e.g HTTP_X_TENANT = microsoft.com resulting regex = (^microsoft\.com$)
            // if the provided value is a comma separated string e.g microsoft.com, bing.com generate (^microsoft\.com$)|(^bing\.com$)
            
            return string.Join("|",  value.Replace(".", SLASH_DOT)
                                    .Split(new char []{','}, StringSplitOptions.None)
                                    .Select(match => string.Format("(^{0}$)", match.Trim())));
        }       
    }
}
